{{ soil.name }}
SoilColor={{ soil.soilcolor }}
SoilGamDry={{ soil.soilgamdry }}
SoilGamWet={{ soil.soilgamwet }}
SoilInitialVoidRatio={{ soil.soilinitialvoidratio }}
SoilCohesion={{ soil.soilcohesion }}
SoilPhi={{ soil.soilphi }}
SoilPreconIsotacheType={{ soil.soilpreconisotachetype.value }}
SoilPreconKoppejanType={{ soil.soilpreconkoppejantype.value }}
SoilUseEquivalentAge={{ soil.soiluseequivalentage.value }}
SoilEquivalentAge={{ soil.soilequivalentage }}
SoilPc={{ soil.soilpc }}
SoilOCR={{ soil.soilocr }}
SoilPOP={{ soil.soilpop }}
SoilLimitStress={{ soil.soillimitstress }}
SoilDrained={{ soil.soildrained.value }}
SoilApAsApproximationByCpCs={{ soil.soilapasapproximationbycpcs.value }}
SoilCv={{ soil.soilcv }}
SoilPermeabilityVer={{ soil.soilpermeabilityver }}
SoilPermeabilityHorFactor={{ soil.soilpermeabilityhorfactor }}
SoilStorageType={{ soil.soilstoragetype.value }}
SoilPermeabilityStrainModulus={{ soil.soilpermeabilitystrainmodulus }}
SoilUseProbDefaults={{ soil.soiluseprobdefaults.value }}
SoilStdGamDry={{ soil.soilstdgamdry }}
SoilStdGamWet={{ soil.soilstdgamwet }}
SoilStdCv={{ soil.soilstdcv }}
SoilStdPc={{ soil.soilstdpc }}
SoilStdPriCompIndex={{ soil.soilstdpricompindex }}
SoilStdSecCompIndex={{ soil.soilstdseccompindex }}
SoilStdSecCompRate={{ soil.soilstdseccomprate }}
SoilStdOCR={{ soil.soilstdocr }}
SoilStdPermeabilityVer={{ soil.soilstdpermeabilityver }}
SoilStdPOP={{ soil.soilstdpop }}
SoilStdPermeabilityHorFactor={{ soil.soilstdpermeabilityhorfactor }}
SoilStdInitialVoidRatio={{ soil.soilstdinitialvoidratio }}
SoilStdPermeabilityStrainModulus={{ soil.soilstdpermeabilitystrainmodulus }}
SoilStdLimitStress={{ soil.soilstdlimitstress }}
SoilStdCp={{ soil.soilstdcp }}
SoilStdCp1={{ soil.soilstdcp1 }}
SoilStdCs={{ soil.soilstdcs }}
SoilStdCs1={{ soil.soilstdcs1 }}
SoilStdAp={{ soil.soilstdap }}
SoilStdAsec={{ soil.soilstdasec }}
SoilStdCar={{ soil.soilstdcar }}
SoilStdCa={{ soil.soilstdca }}
SoilStdRRatio={{ soil.soilstdrratio }}
SoilStdCRatio={{ soil.soilstdcratio }}
SoilStdSRatio={{ soil.soilstdsratio }}
SoilStdCrIndex={{ soil.soilstdcrindex }}
SoilStdCcIndex={{ soil.soilstdccindex }}
SoilStdCswIndex={{ soil.soilstdcswindex }}
SoilDistGamDry={{ soil.soildistgamdry.value }}
SoilDistGamWet={{ soil.soildistgamwet.value }}
SoilDistCv={{ soil.soildistcv.value }}
SoilDistdPc={{ soil.soildistdpc.value }}
SoilDistPriCompIndex={{ soil.soildistpricompindex.value }}
SoilDistSecCompIndex={{ soil.soildistseccompindex.value }}
SoilDistSecCompRate={{ soil.soildistseccomprate.value }}
SoilDistOCR={{ soil.soildistocr.value }}
SoilDistPermeabilityVer={{ soil.soildistpermeabilityver.value }}
SoilDistPOP={{ soil.soildistpop.value }}
SoilDistPermeabilityHorFactor={{ soil.soildistpermeabilityhorfactor.value }}
SoilDistInitialVoidRatio={{ soil.soildistinitialvoidratio.value }}
SoilDistPermeabilityStrainModulus={{ soil.soildistpermeabilitystrainmodulus.value }}
SoilDistLimitStress={{ soil.soildistlimitstress.value }}
SoilDistCp={{ soil.soildistcp.value }}
SoilDistCp1={{ soil.soildistcp1.value }}
SoilDistCs={{ soil.soildistcs.value }}
SoilDistCs1={{ soil.soildistcs1.value }}
SoilDistAp={{ soil.soildistap.value }}
SoilDistAsec={{ soil.soildistasec.value }}
SoilDistCar={{ soil.soildistcar.value }}
SoilDistCa={{ soil.soildistca.value }}
SoilDistRRatio={{ soil.soildistrratio.value }}
SoilDistCRatio={{ soil.soildistcratio.value }}
SoilDistSRatio={{ soil.soildistsratio.value }}
SoilDistCrIndex={{ soil.soildistcrindex.value }}
SoilDistCcIndex={{ soil.soildistccindex.value }}
SoilDistCswIndex={{ soil.soildistcswindex.value }}
SoilCorCpCp1={{ soil.soilcorcpcp1 }}
SoilCorCsCp1={{ soil.soilcorcscp1 }}
SoilCorCs1Cp1={{ soil.soilcorcs1cp1 }}
SoilCorApCp1={{ soil.soilcorapcp1 }}
SoilCorASecCp1={{ soil.soilcoraseccp1 }}
SoilCorCrIndexCcIndex={{ soil.soilcorcrindexccindex }}
SoilCorRRatioCRatio={{ soil.soilcorrratiocratio }}
SoilCorCaCcIndexOrCRatio={{ soil.soilcorcaccindexorcratio }}
SoilCorPriCompIndexSecCompIndex={{ soil.soilcorpricompindexseccompindex }}
SoilCorSecCompRateSecCompIndex={{ soil.soilcorseccomprateseccompindex }}
SoilCp={{ soil.soilcp }}
SoilCp1={{ soil.soilcp1 }}
SoilCs={{ soil.soilcs }}
SoilCs1={{ soil.soilcs1 }}
SoilAp={{ soil.soilap }}
SoilAsec={{ soil.soilasec }}
SoilCar={{ soil.soilcar }}
SoilCa={{ soil.soilca }}
SoilCompRatio={{ soil.soilcompratio.value }}
SoilRRatio={{ soil.soilrratio }}
SoilCRatio={{ soil.soilcratio }}
SoilSRatio={{ soil.soilsratio }}
SoilCrIndex={{ soil.soilcrindex }}
SoilCcIndex={{ soil.soilccindex }}
SoilCswIndex={{ soil.soilcswindex }}
SoilPriCompIndex={{ soil.soilpricompindex }}
SoilSecCompIndex={{ soil.soilseccompindex }}
SoilSecCompRate={{ soil.soilseccomprate }}
SoilHorizontalBehaviourType={{ soil.soilhorizontalbehaviourtype.value }}
SoilElasticity={{ soil.soilelasticity }}
SoilDefaultElasticity={{ soil.soildefaultelasticity.value }}

